@@ -1,11 +1,12 @@
 # -*- coding: utf-8 -*-
 import subprocess
-import time
 import struct
 import os
-import datetime
 import json
+import uuid
+import StringIO
 
+import django
 from django.test import TestCase
 from django.core.urlresolvers import reverse
 from django.contrib.auth.models import User
@@ -13,6 +14,12 @@ from django.http import HttpResponseNotAllowed
 from django.conf import settings
 from django.core import management
 
+try:
+    from django.utils.timezone import now as dt_now
+except ImportError:
+    import datetime
+    dt_now = datetime.datetime.now
+
 from .models import APNService, Device, Notification, NotificationPayloadSizeExceeded
 from .http import JSONResponse
 from .utils import generate_cert_and_pkey
@@ -25,10 +32,11 @@ SSL_SERVER_COMMAND = ('openssl', 's_server', '-accept', '2195', '-cert', TEST_PE
 
 
 class APNServiceTest(TestCase):
-    def setUp(self):
-        self.test_server_proc = subprocess.Popen(SSL_SERVER_COMMAND, stdout=subprocess.PIPE)
-        time.sleep(0.5)  # Wait for test server to be started
+    @classmethod
+    def setUpClass(cls):
+        cls.test_server_proc = subprocess.Popen(SSL_SERVER_COMMAND, stdout=subprocess.PIPE)
 
+    def setUp(self):
         cert, key = generate_cert_and_pkey()
         self.service = APNService.objects.create(name='test-service', hostname='127.0.0.1',
                                                  certificate=cert, private_key=key)
@@ -36,10 +44,6 @@ class APNServiceTest(TestCase):
         self.device = Device.objects.create(token=TOKEN, service=self.service)
         self.notification = Notification.objects.create(message='Test message', service=self.service)
 
-    def test_connection_to_remote_apn_host(self):
-        self.assertTrue(self.service.connect())
-        self.service.disconnect()
-
     def test_invalid_payload_size(self):
         n = Notification(message='.' * 250)
         self.assertRaises(NotificationPayloadSizeExceeded, self.service.pack_message, n.payload, self.device)
@@ -73,12 +77,25 @@ class APNServiceTest(TestCase):
         self.assertFalse(form.is_valid())
         self.assertTrue('passphrase' in form.errors)
 
-    def tearDown(self):
-        self.test_server_proc.kill()
+    def test_pushing_notification_in_chunks(self):
+        devices = []
+        for i in xrange(10):
+            token = uuid.uuid1().get_hex() * 2
+            device = Device.objects.create(token=token, service=self.service)
+            devices.append(device)
+
+        started_at = dt_now()
+        self.service.push_notification_to_devices(self.notification, devices, chunk_size=2)
+        device_count = len(devices)
+        self.assertEquals(device_count,
+                          Device.objects.filter(last_notified_at__gte=started_at).count())
+
+    @classmethod
+    def tearDownClass(cls):
+        cls.test_server_proc.kill()
 
 
 class APITest(TestCase):
-
     def setUp(self):
         self.service = APNService.objects.create(name='sandbox', hostname='gateway.sandbox.push.apple.com')
         self.device_token = TOKEN
@@ -222,9 +239,11 @@ class AuthenticationDecoratorTestAuthBasic(TestCase):
 
 
 class NotificationTest(TestCase):
+    @classmethod
+    def setUpClass(cls):
+        cls.test_server_proc = subprocess.Popen(SSL_SERVER_COMMAND, stdout=subprocess.PIPE)
+
     def setUp(self):
-        self.test_server_proc = subprocess.Popen(SSL_SERVER_COMMAND, stdout=subprocess.PIPE)
-        time.sleep(0.5)
         cert, key = generate_cert_and_pkey()
         self.service = APNService.objects.create(name='service', hostname='127.0.0.1',
                                                  private_key=key, certificate=cert)
@@ -253,11 +272,8 @@ class NotificationTest(TestCase):
         self.assertTrue(self.notification.is_valid_length())
 
     def test_extra_property_not_dict(self):
-        class Dummy():
-            dummy = 1
-        def should_raise():
-            self.notification.extra = Dummy()
-        self.assertRaises(TypeError, should_raise)
+        with self.assertRaises(TypeError):
+            self.notification.extra = 111
 
     def test_extra_property_none(self):
         self.notification.extra = None
@@ -285,14 +301,18 @@ class NotificationTest(TestCase):
         self.assertIsNone(notification.last_sent_at)
         self.assertIsNone(notification.pk)
 
-    def tearDown(self):
-        self.test_server_proc.kill()
+    @classmethod
+    def tearDownClass(cls):
+        cls.test_server_proc.kill()
+
 
 class ManagementCommandPushNotificationTest(TestCase):
+    @classmethod
+    def setUpClass(cls):
+        cls.test_server_proc = subprocess.Popen(SSL_SERVER_COMMAND, stdout=subprocess.PIPE)
+
     def setUp(self):
-        self.started_at = datetime.datetime.now()
-        self.test_server_proc = subprocess.Popen(SSL_SERVER_COMMAND, stdout=subprocess.PIPE)
-        time.sleep(0.5)
+        self.started_at = dt_now()
         cert, key = generate_cert_and_pkey()
         self.service = APNService.objects.create(name='service', hostname='127.0.0.1',
                                                  private_key=key, certificate=cert)
@@ -335,13 +355,25 @@ class ManagementCommandPushNotificationTest(TestCase):
         self.assertTrue(Notification.objects.filter(message=msg, last_sent_at__gt=self.started_at).exists())
         self.assertTrue(self.device in Device.objects.filter(last_notified_at__gt=self.started_at))
 
+    def test_either_message_or_extra_option_required(self):
+        # In Django < 1.5 django.core.management.base.BaseCommand.execute
+        # catches CommandError and raises SystemExit instead.
+        exception = SystemExit if django.VERSION < (1, 5) else management.base.CommandError
+
+        with self.assertRaises(exception):
+            management.call_command('push_ios_notification', service=self.service.pk,
+                                    verbosity=0, stderr=StringIO.StringIO())
+
     def tearDown(self):
         if self.IOS_NOTIFICATIONS_PERSIST_NOTIFICATIONS == 'NotSpecified':
             if hasattr(settings, 'IOS_NOTIFICATIONS_PERSIST_NOTIFICATIONS'):
                 del settings.IOS_NOTIFICATIONS_PERSIST_NOTIFICATIONS
         else:
             settings.IOS_NOTIFICATIONS_PERSIST_NOTIFICATIONS = self.IOS_NOTIFICATIONS_PERSIST_NOTIFICATIONS
-        self.test_server_proc.kill()
+
+    @classmethod
+    def tearDownClass(cls):
+        cls.test_server_proc.kill()
 
 
 class ManagementCommandCallFeedbackService(TestCase):
